#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of an OpenGL plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; A few GLX wraps.
;;; package: (gl glx)

(C-include "gl")

(define-syntax error-if-null
  (syntax-rules ()
    ((_ alien message ...)
     (if (alien-null? alien)
	 (error message ...)))))

(define (with-glx-widget widget thunk)
  (let ((xdisplay (glx-widget-xdisplay widget))
	(xwindow (glx-widget-xwindow widget))
	(glxcontext (glx-widget-glxcontext widget)))
    (if (zero? (C-call "glXMakeCurrent" xdisplay xwindow glxcontext))
	(error "glXMakeCurrent failed"))
    (let ((value (with-gl-context thunk)))
      (if (zero? (C-call "glXMakeCurrent" xdisplay (C-enum "None") 0))
	  (error "glXMakeCurrent NULL failed"))
      value)))

(define (glx:swap-buffers widget)
  (let ((xdisplay (glx-widget-xdisplay widget))
	(xwindow (glx-widget-xwindow widget)))
    (C-call "glXSwapBuffers" xdisplay xwindow)))

(define-class (<glx-widget> (constructor () (width height)))
    (<fix-widget>)

  (xdisplay define standard
	    initializer (lambda () (make-alien '|Display|)))
  (xwindow define standard initial-value #f)
  (glxcontext define standard
	      initializer (lambda () (make-alien '(struct |__GLXcontextRec|)))))

(define-method initialize-instance ((widget <glx-widget>) width height)
  (call-next-method widget width height '())
  (assert-glib-locked '(initialize-instance <glx-widget>))
  (add-glib-cleanup widget (make-glx-widget-cleanup
			    (glx-widget-xdisplay widget)
			    (glx-widget-glxcontext widget)))
  (C-call "gtk_widget_set_double_buffered" (gobject-alien widget) 0))

(define (make-glx-widget-cleanup xdisplay glxcontext)
  (named-lambda (glx-widget-cleanup)
    (cleanup-glx-widget xdisplay glxcontext)))

(define (cleanup-glx-widget xdisplay glxcontext)
  (assert-without-interruption 'cleanup-glx-widget)
  (if (not (alien-null? glxcontext))
      (begin
	(C-call "glXDestroyContext" xdisplay glxcontext)
	(alien-null! glxcontext))))

(define-method gtk-widget-destroy-callback ((widget <glx-widget>))
  (without-interruption
   (lambda ()
     (execute-glib-cleanup widget)))
  (call-next-method widget))

(define-method fix-widget-realize-callback ((widget <glx-widget>))
  (assert-glib-locked '(fix-widget-realize-callback <glx-widget>))
  (let* ((GtkWidget (gobject-alien widget))
	 (parent
	  (C-call "gtk_widget_get_parent_window" (make-alien '|GdkWindow|)
		  GtkWidget)))
    (error-if-null parent "Could not get parent:" widget)

    ;; Create widget GdkWindow.
    (let ((xdisplay (C-call "gdk_window_xdisplay" (glx-widget-xdisplay widget)
			    parent))
	  (screen-num (C-call "gdk_window_screen_num" parent))
	  (attribs (make-attribs
		    `(
		      ;; Singleton attributes.
		      ,(C-enum "GLX_RGBA")
		      ,(C-enum "GLX_DOUBLEBUFFER")

		      ;; Key/value attributes.
		      ,(C-enum "GLX_RED_SIZE") 1
		      ,(C-enum "GLX_GREEN_SIZE") 1
		      ,(C-enum "GLX_BLUE_SIZE") 1
		      ,(C-enum "GLX_DEPTH_SIZE") 1

		      ,(C-enum "None")))))
      (let ((visinfo (C-call "glXChooseVisual" (make-alien '|XVisualInfo|)
			      xdisplay screen-num attribs)))
	(free attribs)
	(let ((gdkvisual (C-call "glx_find_gdkvisual" (make-alien '|GdkVisual|)
				 parent visinfo))
	      (attr (malloc (C-sizeof "GdkWindowAttr") '|GdkWindowAttr|))
	      (GdkWindow (fix-widget-window widget)))
	  (C->= attr "GdkWindowAttr window_type" (C-enum "GDK_WINDOW_CHILD"))
	  (C->= attr "GdkWindowAttr wclass" (C-enum "GDK_INPUT_OUTPUT"))
	  (C->= attr "GdkWindowAttr visual" gdkvisual)
	  (C->= attr "GdkWindowAttr event_mask" (C-enum "GDK_ALL_EVENTS_MASK"))
	  (let ((geometry (fix-widget-geometry widget)))
	    (let ((x (fix-rect-x geometry))
		  (y (fix-rect-y geometry))
		  (width (fix-rect-width geometry))
		  (height (fix-rect-height geometry)))
	      (if x (C->= attr "GdkWindowAttr x" x))
	      (if y (C->= attr "GdkWindowAttr y" y))
	      (C->= attr "GdkWindowAttr width" width)
	      (C->= attr "GdkWindowAttr height" height)
	      (C-call "gdk_window_new" GdkWindow parent attr
		      (bit-ior (if x (C-enum "GDK_WA_X") 0)
			       (if y (C-enum "GDK_WA_Y") 0)
			       (C-enum "GDK_WA_VISUAL")))))
	  (free attr)
	  (error-if-null GdkWindow "Could not create GdkWindow:" widget)
	  (set-glx-widget-xwindow! widget (C-call "gdk_window_xid" GdkWindow))
	  (C-call "gtk_widget_set_window" GtkWidget GdkWindow)
	  (C-call "gtk_widget_register_window" GtkWidget GdkWindow)
	  ;; gtk_widget_unregister_window (as well as gdk_window_destroy)
	  ;; are called by the default unrealize method.
	  (let ((alien (glx-widget-glxcontext widget)))
	    (C-call "glXCreateContext" alien xdisplay visinfo 0 1)
	    (error-if-null alien "Could not create GLXContext.")))))

    (C-call "gtk_widget_set_can_focus" GtkWidget 1)))

(define (glx-widget-realized? widget)
  (not (alien-null? (glx-widget-glxcontext widget))))

(declare (integrate-operator bit-ior))
(define (bit-ior . ints)
  (reduce bitwise-ior 0 ints))

#;(define (choose-fb-config display screen-num attrib-list)
  (let ((configs (make-alien '|GLXFBConfig|))
	(copy (make-alien '|GLXFBConfig|))
	(attribs (make-attribs attrib-list))
	(num-configs (malloc (C-sizeof "int") 'int)))
    (add-glib-cleanup configs (make-fb-configs-cleanup copy))
    (C-call "glXChooseFBConfig" copy
	    display screen-num attribs num-configs)
    (free attribs)
    (free num-configs)
    (copy-alien-address! configs copy)
    (error-if-null configs "Could not find:" configs)
    configs))

#;(define (make-fb-configs-cleanup alien)
  (named-lambda (fb-configs-cleanup)
    (cleanup-fb-configs alien)))

#;(define (cleanup-fb-configs alien)
  (assert-without-interrupts 'cleanup-fb-configs)
  (if (not (alien-null? alien))
      (begin
	(C-call "XFree" alien)
	(alien-null! alien))))

(define (make-attribs attribs)
  (let* ((len (length attribs))
	 (alien (malloc (* len (C-sizeof "int")) '|int|)))
    (do ((attribs attribs (cdr attribs))
	 (i 0 (fix:1+ i)))
	((fix:= i len))
      (let ((attrib (car attribs)))
	(guarantee integer? attrib 'make-attribs)
	((ucode-primitive c-poke-int 3)
	 alien
	 (fix:* i (C-sizeof "int"))
	 attrib)))
    alien))

#;(define (glx-widget-write-to-png widget filename)
  (let* ((pixbuf (make-pixbuf))
	 (alien (gobject-alien pixbuf))
	 (geometry (fix-widget-geometry widget)))
    (C-call "gdk_pixbuf_xlib_get_from_drawable" alien
	    0 (glx-widget-xwindow widget) ;dest, src
	    0 0				;cmap, visual
	    0 0				;src_x, src_y
	    0 0				;dest_x, dest_y
	    (fix-rect-width geometry)
	    (fix-rect-height geometry))
    (let ((gerror* (make-gerror-pointer)))
      (if (zero? (C-call "gdk_pixbuf_save" alien filename "png" gerror* 0 0 0))
	  (let* ((gerror (C-> gerror* "* GError"))
		 (message
		  (or (and (not (alien-null? gerror))
			   (utf8->string
			    (c-peek-cstring (C-> gerror "GError message"))))
		      "GError pointer not set.")))
	    (gerror-pointer-free gerror*)
	    (error message))))))

(define-class (<glx-viewport> (constructor () (width height)))
    ;; A <glx-widget> with camera parameters, and a default key-press
    ;; handler to fly the camera.
    (<glx-widget>)
  (position define accessor initializer (lambda () (flo:3d 0. 0. 0.)))
  (heading define standard initial-value 0.)
  (tilt define standard initial-value 0.)
  (step-size define standard initial-value 1.)
  (fovy define standard initial-value 45.)
  (near define standard initial-value .1)
  (far define standard initial-value 1000.))

(define-method initialize-instance ((widget <glx-viewport>) width height)
  (call-next-method widget width height)
  (set-gtk-widget-draw-callback! widget draw-callback)
  (set-fix-widget-key-press-handler! widget glx-viewport-key-press-handler))

(define-method fix-widget-new-geometry-callback ((widget <glx-viewport>))
  (call-next-method widget)
  (if (glx-widget-xwindow widget)
      (let ((geometry (fix-widget-geometry widget))
	    (scale (C-call "gdk_window_get_scale_factor"
			   (fix-widget-window widget))))
	(let ((w (fix:* scale (fix-rect-width geometry)))
	      (h (fix:* scale (fix-rect-height geometry))))
	  (let ((aspect (->flonum (/ w h))))
	    (with-glx-widget widget
	      (lambda ()
		(gl:matrix-mode 'projection)
		(gl:load-identity)
		(gl:viewport 0 0 w h)
		(glu:perspective (glx-viewport-fovy widget)
				 aspect
				 (glx-viewport-near widget)
				 (glx-viewport-far widget)))))))))

(define (draw-callback widget cr)
  (declare (ignore cr))
  #;(%trace "; draw "widget" at "
	   (cairo-clip-extents
	    cr (lambda (min-x min-y max-x max-y)
		 (define-integrable n->s number->string)
		 (string-append (n->s min-x)","(n->s min-y)
				" "(n->s (- max-x min-x))
				"x"(n->s (- max-y min-y)))))
	   "\n")
  (with-glx-widget widget
    (lambda ()
      (gl:matrix-mode 'modelview)
      (gl:load-identity)
      (let ((position (glx-viewport-position widget))
	    ;; Heading 0.    is N: ( 0. 0. -1.)
	    ;; Heading pi/2  is E: ( 1. 0.  0.)
	    ;; Heading pi    is S: ( 0. 0.  1.)
	    ;; Heading 3pi/2 is W: (-1. 0.  0.)
	    ;; Heading h       is: ((sin h) 0. (- (cos h)))
	    (heading (glx-viewport-heading widget))
	    (tilt (glx-viewport-tilt widget)))
	(let ((sin-heading (flo:sin heading))
	      (-cos-heading (flo:- 0. (flo:cos heading)))
	      (sin-tilt (flo:sin tilt))
	      (cos-tilt (flo:cos tilt)))
	  (let ((center (flo:3d (flo:+ (x position) sin-heading)
				(flo:+ (y position) sin-tilt)
				(flo:+ (z position) -cos-heading)))
		(up     (flo:3d (flo:* sin-tilt sin-heading)
				cos-tilt
				(flo:* sin-tilt -cos-heading))))
	    (glu:look-at position center up))))
      (gl:clear 'color-buffer 'depth-buffer)
      (glx-viewport-draw widget)
      (glx:swap-buffers widget)
      #t)))

(define-generic glx-viewport-draw (widget))

(define-generic glx-viewport-key-press-handler (widget key bits))

(define-method glx-viewport-key-press-handler ((widget <glx-viewport>) key bits)
  (%trace "; (key-press <glx-viewport>) "widget" "key" "bits"\n")
  (case key
    ((#\escape #\Q #\q)
     (gtk-widget-destroy (let loop ((w widget))
			   (let ((p (gtk-widget-parent w)))
			     (if p
				 (loop p)
				 w))))
     #t)
    ((#\? #\h #\H #\c-h |F1|)
     (popup-help widget)
     #t)
    ((|Up|)
     (cond ((fix:= bits char-bit:control) (tilt! widget 5.))
	   ((fix:= bits 0) (step! widget 1.))
	   (else #f)))
    ((|Down|)
     (cond ((fix:= bits char-bit:control) (tilt! widget -5.))
	   ((fix:= bits 0) (step! widget -1.))
	   (else #f)))
    ((|Right|)
     (turn! widget 5.))
    ((|Left|)
     (turn! widget -5.))
    (else
     #f)))

(define (turn! widget degrees)
  (let ((new-heading (flo:+ (glx-viewport-heading widget)
			    (degrees->radians degrees))))
    (set-glx-viewport-heading!
     widget (cond ((flo:> new-heading pi) (flo:- new-heading 2pi))
		  ((flo:< new-heading -pi) (flo:+ new-heading 2pi))
		  (else new-heading))))
  (gtk-widget-queue-draw widget)
  #t)

(define (tilt! widget degrees)
  (let ((tilt (flo:+ (glx-viewport-tilt widget)
		     (degrees->radians degrees))))
    (if (and (flo:< -pi/2 tilt) (flo:< tilt pi/2))
	(begin
	  (set-glx-viewport-tilt! widget tilt)
	  (gtk-widget-queue-draw widget)
	  #t)
	#f)))

(define (step! widget number)
  (let ((distance (flo:* number (glx-viewport-step-size widget)))
	(position (glx-viewport-position widget))
	(heading (glx-viewport-heading widget))
	(tilt (glx-viewport-tilt widget)))
    (let ((sin-heading (flo:sin heading))
	  (cos-heading (flo:cos heading))
	  (sin-tilt (flo:sin tilt))
	  (cos-tilt (flo:cos tilt)))
      (set-x! position (flo:+ (x position)
			      (flo:* distance (flo:* cos-tilt sin-heading))))
      (set-y! position (flo:+ (y position)
			      (flo:* distance sin-tilt)))
      (set-z! position (flo:- (z position)
			      (flo:* distance (flo:* cos-tilt cos-heading))))))
  (gtk-widget-queue-draw widget)
  #t)

(define help-window #f)

(define (popup-help widget)
  (declare (ignore widget))
  (if (or (not help-window)
	  (gtk-widget-destroyed? help-window))
      (let ((window (gtk-window-new 'toplevel))
	    (label (gtk-label-new
		    "Key - Viewport command
ESC - Quit.
Up/Down - Move forward/backward.
Left/Right - Turn left/right.
Ctrl-Up/Down - Tilt up/down.")))
	(gtk-window-set-type-hint window 'utility)
	(gtk-window-set-accept-focus window #f)
	(gtk-window-set-title window "Help")
	(gtk-container-set-border-width window 20)
	(gtk-container-add window label)
	(set! help-window window)
	(gtk-widget-show-all window)))
  (gtk-window-present help-window))

(define-integrable (flo:3d x y z)
  (let ((v (flo:vector-cons 3)))
    (flo:vector-set! v 0 x)
    (flo:vector-set! v 1 y)
    (flo:vector-set! v 2 z)
    v))

(define-integrable (x v) (flo:vector-ref v 0))
(define-integrable (y v) (flo:vector-ref v 1))
(define-integrable (z v) (flo:vector-ref v 2))
(define-integrable (set-x! v f) (flo:vector-set! v 0 f))
(define-integrable (set-y! v f) (flo:vector-set! v 1 f))
(define-integrable (set-z! v f) (flo:vector-set! v 2 f))

(define-integrable pi/4 (flo:atan2 1. 1.))
(define-integrable pi/2 (flo:* 2. pi/4))
(define-integrable pi (flo:* 4. pi/4))
(define-integrable 2pi (flo:* 8. pi/4))
(define-integrable -pi (flo:- 0. pi))
(define-integrable -pi/2 (flo:- 0. pi/2))

(define-integrable (degrees->radians degrees) (flo:* degrees (flo:/ 2pi 360.)))

(define-integrable (radians->degrees radians) (flo:* radians (flo:/ 360. 2pi)))

(define %trace? #f)

(define-syntax %trace
  (syntax-rules ()
    ((_ args ...)
     (if %trace? (outf-error args ...)))))